@@ -59,7 +59,7 @@ class TokenAuthentication(Authentication):
         self.server = server
         self.skip_tls = skip_tls
 
-    def login(self):
+    def login(self) -> str:
         """
         This function is used to login to an OpenShift cluster using the user's API token and API server address.
         Depending on the cluster, a user can choose to login in with "--insecure-skip-tls-verify` by setting `skip_tls`
@@ -74,15 +74,18 @@ class TokenAuthentication(Authentication):
             error_msg = osp.result.err()
             if "The server uses a certificate signed by unknown authority" in error_msg:
                 return "Error: certificate auth failure, please set `skip_tls=True` in TokenAuthentication"
+            elif "invalid" in error_msg:
+                raise PermissionError(error_msg)
             else:
                 return error_msg
         return response.out()
 
-    def logout(self):
+    def logout(self) -> str:
         """
         This function is used to logout of an OpenShift cluster.
         """
-        response = oc.invoke("logout")
+        args = [f"--token={self.token}", f"--server={self.server}:6443"]
+        response = oc.invoke("logout", args)
         return response.out()
 
 
@@ -104,14 +107,14 @@ class PasswordUserAuthentication(Authentication):
         self.username = username
         self.password = password
 
-    def login(self):
+    def login(self) -> str:
         """
         This function is used to login to an OpenShift cluster using the user's `username` and `password`.
         """
         response = oc.login(self.username, self.password)
         return response.out()
 
-    def logout(self):
+    def logout(self) -> str:
         """
         This function is used to logout of an OpenShift cluster.
         """
